require "postfix/rate/policyd/agent_backend"

RSpec.describe Postfix::Rate::Policyd::DataBackend do
  before :each do
    DatabaseHelpers.init_database
  end

  subject do
    Postfix::Rate::Policyd::DataBackend.new
  end

  let :anon_request do
    req = Postfix::Rate::Policyd::Request.new
    req.set_attribute 'instance', '123.456.789'
    req.set_attribute 'sender', 'sender@foo.bar'
    req.set_attribute 'recipient', 'recipient@foo.bar'
    req.set_attribute 'client_address', '1.2.3.4'
    req.set_attribute 'size', 1000
    req.set_attribute 'helo_name', 'foo.bar'
    req.set_attribute 'recipient_count', 5
    req.new_instance!
    req
  end

  let :auth_request do
    req = Postfix::Rate::Policyd::Request.new
    req.set_attribute 'instance', '123.456.789'
    req.set_attribute 'sender', 'sender@foo.bar'
    req.set_attribute 'recipient', 'recipient@foo.bar'
    req.set_attribute 'sasl_username', 'sender@foo.bar'
    req.set_attribute 'client_address', '1.2.3.4'
    req.set_attribute 'size', 1000
    req.set_attribute 'helo_name', 'foo.bar'
    req.set_attribute 'recipient_count', 5
    req.new_instance!
    req
  end

  it 'should create an anon identity on fetch' do
    expect {
      expect(entity = subject.fetch_entity(anon_request)).to be_a Postfix::Rate::Models::ClientAddress
      expect(entity.policy).to be_a Postfix::Rate::Models::Policy
      expect(entity.policy.name).to eq 'default_for_addresses'
      expect(entity.tracker).to be_a Postfix::Rate::Models::Tracker
    }.to change{Postfix::Rate::Models::ClientAddress.count}.by(1)
      .and change{Postfix::Rate::Models::Tracker.count}.by(1)
  end

  it 'should create an auth identity on fetch' do
    expect {
      expect(entity = subject.fetch_entity(auth_request)).to be_a Postfix::Rate::Models::AuthenticatedIdentity
      expect(entity.policy).to be_a Postfix::Rate::Models::Policy
      expect(entity.policy.name).to eq 'default_for_identities'
      expect(entity.tracker).to be_a Postfix::Rate::Models::Tracker
    }.to change{Postfix::Rate::Models::AuthenticatedIdentity.count}.by(1)
      .and change{Postfix::Rate::Models::Tracker.count}.by(1)
  end

  it 'should fetch an existing anon identity' do
    Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
    expect {
      expect(subject.fetch_entity(anon_request)).to be_a Postfix::Rate::Models::ClientAddress
    }.not_to change{Postfix::Rate::Models::ClientAddress.count}
  end

  it 'should fetch an existing auth identity' do
    Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
    expect {
      expect(subject.fetch_entity(auth_request)).to be_a Postfix::Rate::Models::AuthenticatedIdentity
    }.not_to change{Postfix::Rate::Models::AuthenticatedIdentity.count}
  end

  context 'with anon request' do
    context 'check_initial_quota' do
      it 'should return success on first hit' do
        expect(subject.check_initial_quota(anon_request)).to be_ack
      end

      it 'should fail if the identity is disabled' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.disable! 'foo bar'

        expect(subject.check_initial_quota(anon_request)).to be_nack
      end

      it 'should fail if the address is blacklisted' do
        Postfix::Rate::Models::NetworkAddress.create(address: '1.2.3.4', notes: 'salutava sempre..', inserted: Time.now)

        expect(subject.check_initial_quota(anon_request)).to be_nack
      end

      it 'should fail if the count tracker is over' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.tracker.count = id.policy.count
        id.tracker.save

        expect(subject.check_initial_quota(anon_request)).to be_nack
      end
    end

    context 'check_rcpt_quota' do
      it 'should return success on first hit' do
        expect(subject.check_rcpt_quota(anon_request)).to be_ack
      end

      it 'should fail if the recipients tracker is over' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.policy.recipients = 100
        id.policy.save
        id.tracker.recipients = id.policy.recipients
        id.tracker.save

        expect(subject.check_rcpt_quota(anon_request)).to be_nack
      end
    end

    context 'check_final_quota_and_update' do
      it 'should return success on first hit' do
        expect(subject.check_final_quota_and_update(anon_request)).to be_ack
      end

      it 'should fail if the volume tracker is over' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.policy.volume = 100
        id.policy.save
        id.tracker.volume = id.policy.volume
        id.tracker.save

        expect(subject.check_final_quota_and_update(anon_request)).to be_nack
      end

      it 'should update the count tracker' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.tracker.count = 1
        id.tracker.save

        expect {
          expect(subject.check_final_quota_and_update(anon_request)).to be_ack
          id.tracker.reload
        }.to change{id.tracker.count}.by(1)
      end

      it 'should decrement the count tracker and add the new count' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.tracker.count = 100
        id.tracker.save

        # half interval
        # (current quota / 2) + 1

        Timecop.travel((id.policy.interval/2).seconds.from_now) do
          expect {
            expect(subject.check_final_quota_and_update(anon_request)).to be_ack
            id.tracker.reload
          }.to change{id.tracker.count}.to(51)
        end
      end

      it 'should reset the count tracker' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.tracker.count = 100
        id.tracker.save

        Timecop.travel((id.policy.interval + 10).seconds.from_now) do
          expect {
            expect(subject.check_final_quota_and_update(anon_request)).to be_ack
            id.tracker.reload
          }.to change{id.tracker.count}.to(1)
        end
      end

      it 'should update the recipients tracker' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.tracker.recipients = 1
        id.tracker.save
        
        expect {
          expect(subject.check_final_quota_and_update(anon_request)).to be_ack
          id.tracker.reload
        }.to change{id.tracker.recipients}.by(5)
      end

      it 'should decrement the recipients tracker and add the current count' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.tracker.recipients = 100
        id.tracker.save

        # half interval
        # (current quota / 2) + 5

        Timecop.travel((id.policy.interval/2).seconds.from_now) do
          expect {
            expect(subject.check_final_quota_and_update(anon_request)).to be_ack
            id.tracker.reload
          }.to change{id.tracker.recipients}.to(55)
        end
      end

      it 'should reset the recipients tracker' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.tracker.recipients = 100
        id.tracker.save

        Timecop.travel((id.policy.interval + 10).seconds.from_now) do
          expect {
            expect(subject.check_final_quota_and_update(anon_request)).to be_ack
            id.tracker.reload
          }.to change{id.tracker.recipients}.to(5)
        end
      end

      it 'should update the volume tracker' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.tracker.volume = 0
        id.tracker.save
        
        expect {
          expect(subject.check_final_quota_and_update(anon_request)).to be_ack
          id.tracker.reload
        }.to change{id.tracker.volume}.by(1000)
      end

      it 'should decrement the volume tracker and add the current count' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.tracker.volume = 10000
        id.tracker.save

        # half interval
        # (current quota / 2) + 1000

        Timecop.travel((id.policy.interval/2).seconds.from_now) do
          expect {
            expect(subject.check_final_quota_and_update(anon_request)).to be_ack
            id.tracker.reload
          }.to change{id.tracker.volume}.to(6000)
        end
      end

      it 'should reset the volume tracker' do
        id = Postfix::Rate::Models::ClientAddress.find_or_create(address: '1.2.3.4')
        id.tracker.volume = 10000
        id.tracker.save

        Timecop.travel((id.policy.interval + 10).seconds.from_now) do
          expect {
            expect(subject.check_final_quota_and_update(anon_request)).to be_ack
            id.tracker.reload
          }.to change{id.tracker.volume}.to(1000)
        end
      end
    end
  end

  context 'with auth request' do
    context 'check_initial_quota' do
      it 'should return success on first hit' do
        expect(subject.check_initial_quota(auth_request)).to be_ack
      end

      it 'should fail if the identity is disabled' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.disable! 'foo bar'

        expect(subject.check_initial_quota(auth_request)).to be_nack
      end

      it 'should fail if the address is blacklisted' do
        Postfix::Rate::Models::NetworkAddress.create(address: '1.2.3.4', notes: 'salutava sempre..', inserted: Time.now)

        expect(subject.check_initial_quota(auth_request)).to be_nack
      end

      it 'should fail if the count tracker is over' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.tracker.count = id.policy.count
        id.tracker.save

        expect(subject.check_initial_quota(auth_request)).to be_nack
      end
    end

    context 'check_rcpt_quota' do
      it 'should return success on first hit' do
        expect(subject.check_rcpt_quota(auth_request)).to be_ack
      end

      it 'should fail if the recipients tracker is over' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.policy.recipients = 100
        id.policy.save
        id.tracker.recipients = id.policy.recipients
        id.tracker.save

        expect(subject.check_rcpt_quota(auth_request)).to be_nack
      end
    end

    context 'check_final_quota_and_update' do
      it 'should return success on first hit' do
        expect(subject.check_final_quota_and_update(auth_request)).to be_ack
      end

      it 'should fail if the volume tracker is over' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.policy.volume = 100
        id.policy.save
        id.tracker.volume = id.policy.volume
        id.tracker.save

        expect(subject.check_final_quota_and_update(auth_request)).to be_nack
      end

      it 'should update the count tracker' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.tracker.count = 1
        id.tracker.save

        expect {
          expect(subject.check_final_quota_and_update(auth_request)).to be_ack
          id.tracker.reload
        }.to change{id.tracker.count}.by(1)
      end

      it 'should decrement the count tracker and add the new count' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.tracker.count = 100
        id.tracker.save

        # half interval
        # (current quota / 2) + 1

        Timecop.travel((id.policy.interval/2).seconds.from_now) do
          expect {
            expect(subject.check_final_quota_and_update(auth_request)).to be_ack
            id.tracker.reload
          }.to change{id.tracker.count}.to(51)
        end
      end

      it 'should reset the count tracker' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.tracker.count = 100
        id.tracker.save

        Timecop.travel((id.policy.interval + 10).seconds.from_now) do
          expect {
            expect(subject.check_final_quota_and_update(auth_request)).to be_ack
            id.tracker.reload
          }.to change{id.tracker.count}.to(1)
        end
      end

      it 'should update the recipients tracker' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.tracker.recipients = 1
        id.tracker.save
        
        expect {
          expect(subject.check_final_quota_and_update(auth_request)).to be_ack
          id.tracker.reload
        }.to change{id.tracker.recipients}.by(5)
      end

      it 'should decrement the recipients tracker and add the current count' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.tracker.recipients = 100
        id.tracker.save

        # half interval
        # (current quota / 2) + 5

        Timecop.travel((id.policy.interval/2).seconds.from_now) do
          expect {
            expect(subject.check_final_quota_and_update(auth_request)).to be_ack
            id.tracker.reload
          }.to change{id.tracker.recipients}.to(55)
        end
      end

      it 'should reset the recipients tracker' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.tracker.recipients = 100
        id.tracker.save

        Timecop.travel((id.policy.interval + 10).seconds.from_now) do
          expect {
            expect(subject.check_final_quota_and_update(auth_request)).to be_ack
            id.tracker.reload
          }.to change{id.tracker.recipients}.to(5)
        end
      end

      it 'should update the volume tracker' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.tracker.volume = 0
        id.tracker.save
        
        expect {
          expect(subject.check_final_quota_and_update(auth_request)).to be_ack
          id.tracker.reload
        }.to change{id.tracker.volume}.by(1000)
      end

      it 'should decrement the volume tracker and add the current count' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.tracker.volume = 10000
        id.tracker.save

        # half interval
        # (current quota / 2) + 1000

        Timecop.travel((id.policy.interval/2).seconds.from_now) do
          expect {
            expect(subject.check_final_quota_and_update(auth_request)).to be_ack
            id.tracker.reload
          }.to change{id.tracker.volume}.to(6000)
        end
      end

      it 'should reset the volume tracker' do
        id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'sender@foo.bar')
        id.tracker.volume = 10000
        id.tracker.save

        Timecop.travel((id.policy.interval + 10).seconds.from_now) do
          expect {
            expect(subject.check_final_quota_and_update(auth_request)).to be_ack
            id.tracker.reload
          }.to change{id.tracker.volume}.to(1000)
        end
      end
    end
  end
end