require "postfix/rate/policyd/agent_backend"

RSpec.describe Postfix::Rate::Policyd::Heuristics do
  before :each do
    DatabaseHelpers.init_database
  end

  subject do
    Postfix::Rate::Policyd::DataBackend.new
  end

  def request
    req = Postfix::Rate::Policyd::Request.new
    req.set_attribute 'instance', '123.456.789'
    req.set_attribute 'sender', 'sender@foo.bar'
    req.set_attribute 'recipient', 'recipient@foo.bar'
    req.set_attribute 'sasl_username', 'sender@foo.bar'
    req.set_attribute 'client_address', '1.2.3.4'
    req.set_attribute 'size', 1000
    req.set_attribute 'helo_name', 'foo.bar'
    req.set_attribute 'recipient_count', 5
    req.new_instance!
    req
  end

  it 'should initialize the heuristics data' do
    req = request
    entity = subject.fetch_entity(req)

    expect {
      subject.check_heuristics(req)
    }.to change{entity.tracker.heuristics}
  end

  it 'should add the new address in the heuristics data' do
    req = request
    entity = subject.fetch_entity(req)
    subject.update_quota(req)

    req1 = request
    req1.set_attribute 'client_address', '2.3.4.5'
    req1.new_instance!
    entity = subject.fetch_entity(req1)

    expect {
      subject.check_heuristics(req1)
    }.to change{entity.tracker.heuristics[:addresses].count}.to(2)
  end

  it 'should update the address data' do
    req = request
    entity = subject.fetch_entity(req)
    subject.update_quota(req)

    Timecop.travel(10.seconds.from_now) do
      req1 = request
      entity = subject.fetch_entity(req1)

      expect {
        subject.check_heuristics(req1)
      }.to change{entity.tracker.heuristics[:addresses]['1.2.3.4'][:last_seen]}
        .and change{entity.tracker.heuristics[:addresses]['1.2.3.4'][:total_count]}
        .and change{entity.tracker.heuristics[:addresses]['1.2.3.4'][:last_count]}
    end
  end

  it 'last_count is a rolling average' do
    req = request
    entity = subject.fetch_entity(req)
    subject.update_quota(req)

    Timecop.travel((entity.policy.h_interval/2).seconds.from_now) do
      req1 = request
      entity = subject.fetch_entity(req1)

      expect {
        subject.check_heuristics(req1)
      }.to change{entity.tracker.heuristics[:addresses]['1.2.3.4'][:last_count]}.to(1.5)
    end
  end

  it 'should reset last_count' do
    req = request
    entity = subject.fetch_entity(req)
    subject.update_quota(req)

    Timecop.travel((entity.policy.h_interval+10).seconds.from_now) do
      req1 = request
      entity = subject.fetch_entity(req1)
      
      entity.tracker.heuristics[:addresses]['1.2.3.4'][:last_count] = 100

      expect {
        subject.check_heuristics(req1)
      }.to change{entity.tracker.heuristics[:addresses]['1.2.3.4'][:last_count]}.to(1)
    end
  end

  context 'address blacklisting' do   
    it 'should never blacklist and ban clients from a single address' do
      req = request
      entity = subject.fetch_entity(req)
      subject.update_quota(req)

      Timecop.travel(10.seconds.from_now) do
        req1 = request
        entity = subject.fetch_entity(req1)
        
        entity.tracker.heuristics[:addresses]['1.2.3.4'][:last_count] = entity.policy.h_interval_count*2
        entity.tracker.heuristics[:addresses]['1.2.3.4'][:total_count] = entity.policy.h_interval_count*2

        expect {
          subject.check_heuristics(req1)
        }.to not_change{entity.disabled}
          .and not_change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}
      end
    end  

    it 'should blacklist clients with multiple addresses and one address over quota' do
      req = request
      subject.fetch_entity(req)
      subject.update_quota(req)
      
      req1 = request
      req1.set_attribute 'client_address', '2.3.4.5'
      req1.new_instance!
      subject.fetch_entity(req1)
      subject.update_quota(req1)

      req2 = request
      req2.set_attribute 'client_address', '2.3.4.5'
      req2.new_instance!
      entity = subject.fetch_entity(req2)
      
      entity.tracker.heuristics[:addresses]['2.3.4.5'][:last_count] = entity.policy.h_interval_count
      entity.tracker.heuristics[:addresses]['2.3.4.5'][:total_count] = entity.policy.h_interval_count

      Timecop.travel(10.seconds.from_now) do
        expect {
          subject.check_heuristics(req2)
        }.to change{entity.disabled}.to(true)
          .and change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}.by(1)
      end
    end

    it 'should respect the rolling average' do
      req = request
      subject.fetch_entity(req)
      subject.update_quota(req)
      
      req1 = request
      req1.set_attribute 'client_address', '2.3.4.5'
      req1.new_instance!
      subject.fetch_entity(req1)
      subject.update_quota(req1)

      req2 = request
      req2.set_attribute 'client_address', '2.3.4.5'
      req2.new_instance!
      entity = subject.fetch_entity(req2)
      
      entity.tracker.heuristics[:addresses]['2.3.4.5'][:last_count] = entity.policy.h_interval_count
      entity.tracker.heuristics[:addresses]['2.3.4.5'][:total_count] = entity.policy.h_interval_count

      Timecop.travel((entity.policy.h_interval/2).seconds.from_now) do
        expect {
          subject.check_heuristics(req2)
        }.to not_change{entity.disabled}
          .and not_change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}
      end
    end

    it 'should respect the heuristic interval' do
      req = request
      subject.fetch_entity(req)
      subject.update_quota(req)
      
      req1 = request
      req1.set_attribute 'client_address', '2.3.4.5'
      req1.new_instance!
      subject.fetch_entity(req1)
      subject.update_quota(req1)

      req2 = request
      req2.set_attribute 'client_address', '2.3.4.5'
      req2.new_instance!
      entity = subject.fetch_entity(req2)
      
      entity.tracker.heuristics[:addresses]['2.3.4.5'][:last_count] = entity.policy.h_interval_count
      entity.tracker.heuristics[:addresses]['2.3.4.5'][:total_count] = entity.policy.h_interval_count

      Timecop.travel((entity.policy.h_interval+10).seconds.from_now) do
        expect {
          subject.check_heuristics(req2)
        }.to not_change{entity.disabled}
          .and not_change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}
      end
    end

    it 'should not blacklist addresses after a grace period' do
      req = request
      subject.fetch_entity(req)
      subject.update_quota(req)
      
      req1 = request
      req1.set_attribute 'client_address', '2.3.4.5'
      req1.new_instance!
      entity = subject.fetch_entity(req1)
      subject.update_quota(req1)
      
      Timecop.travel((entity.policy.h_grace_threshold-10).seconds.from_now) do
        req2 = request
        req2.set_attribute 'client_address', '2.3.4.5'
        req2.new_instance!
        subject.fetch_entity(req2)
        subject.update_quota(req2)
      end
      
      Timecop.travel((entity.policy.h_grace_threshold+10).seconds.from_now) do
        req3 = request
        req3.set_attribute 'client_address', '2.3.4.5'
        req3.new_instance!
        entity = subject.fetch_entity(req3)
        entity.tracker.heuristics[:addresses]['2.3.4.5'][:last_count] = entity.policy.h_interval_count*2
        entity.tracker.heuristics[:addresses]['2.3.4.5'][:total_count] = entity.policy.h_interval_count*2

        expect {
          subject.check_heuristics(req3)
        }.to not_change{entity.disabled}
          .and not_change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}
      end
    end

    it 'should ban clients sending from too many addresses' do
      req = request
      entity = subject.fetch_entity(req)

      # build and store the maximum allowed amount of addresses
      addresses = 1.upto(entity.policy.h_max_addresses).map do | n |
        "1.1.1.#{n}"
      end

      addresses.each.with_index do | addr, idx |
        req = request
        req.set_attribute 'client_address', addr
        req.new_instance!
        entity = subject.fetch_entity(req)
        subject.update_quota(req)
      end

      # now attempt a request from a new address
      req = request
      req.set_attribute 'client_address', '2.2.2.2'
      req.new_instance!
      entity = subject.fetch_entity(req)
      
      # should forget 1.1.1.1 and store 2.2.2.2
      new_addresses = addresses.dup
      new_addresses.shift
      new_addresses.push('2.2.2.2')

      expect {
        subject.check_heuristics(req)
      }.to change{entity.disabled}
        .and change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}
        .and not_change{entity.tracker.heuristics[:addresses].keys}
    end

    it 'should expire old addresses' do
      req = request
      entity = subject.fetch_entity(req)

      # build and store the maximum allowed amount of addresses
      addresses = 1.upto(entity.policy.h_max_addresses).map do | n |
        "1.1.1.#{n}"
      end

      addresses.each.with_index do | addr, idx |
        interval = entity.policy.h_activity_threshold / entity.policy.h_max_addresses
        
        Timecop.travel((interval*idx).seconds.from_now) do
          req = request
          req.set_attribute 'client_address', addr
          req.new_instance!
          entity = subject.fetch_entity(req)
          subject.update_quota(req)
        end
      end

      # now attempt a request from a new address
      Timecop.travel((entity.policy.h_activity_threshold+10).seconds.from_now) do
        req = request
        req.set_attribute 'client_address', '2.2.2.2'
        req.new_instance!
        entity = subject.fetch_entity(req)
        
        # should forget 1.1.1.1 and store 2.2.2.2
        new_addresses = addresses.dup
        new_addresses.shift
        new_addresses.push('2.2.2.2')

        expect {
          subject.check_heuristics(req)
        }.to not_change{entity.disabled}
          .and not_change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}
          .and change{entity.tracker.heuristics[:addresses].keys}
            .from(addresses).to(new_addresses)
      end
    end
  end

  context 'address whitelist' do 
    it 'should allow withelisted address when over quota' do
      Postfix::Rate::Models::NetworkAddress.create(address: '2.3.4.5', notes: 'salutava sempre...', inserted: Time.now, flags: 'w')

      req = request
      subject.fetch_entity(req)
      subject.update_quota(req)
      
      req1 = request
      req1.set_attribute 'client_address', '2.3.4.5'
      req1.new_instance!
      subject.fetch_entity(req1)
      subject.update_quota(req1)

      req2 = request
      req2.set_attribute 'client_address', '2.3.4.5'
      req2.new_instance!
      entity = subject.fetch_entity(req2)
      subject.check_blacklisting(req2) # this will load the address whitelist
      
      # this normally should result in a ban
      entity.tracker.heuristics[:addresses]['2.3.4.5'][:last_count] = entity.policy.h_interval_count
      entity.tracker.heuristics[:addresses]['2.3.4.5'][:total_count] = entity.policy.h_interval_count
      entity.tracker.save

      Timecop.travel(10.seconds.from_now) do
        expect {          
          subject.update_quota(req2)
        }.to not_change{entity.disabled}
          .and not_change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}
      end
    end

    it 'should allow whitelisted address even when address count is over the maximum allowed' do
      Postfix::Rate::Models::NetworkAddress.create(address: '2.3.4.5', notes: 'salutava sempre...', inserted: Time.now, flags: 'w')
      
      req = request
      entity = subject.fetch_entity(req)

      # build and store the maximum allowed amount of addresses
      addresses = 1.upto(entity.policy.h_max_addresses).map do | n |
        "1.1.1.#{n}"
      end

      addresses.each.with_index do | addr, idx |
        req = request
        req.set_attribute 'client_address', addr
        req.new_instance!
        entity = subject.fetch_entity(req)
        subject.update_quota(req)
      end

      # now attempt a request from a new address
      req = request
      req.set_attribute 'client_address', '2.3.4.5'
      req.new_instance!
      entity = subject.fetch_entity(req)
      subject.check_blacklisting(req) # this will load the address whitelist
      
      expect {
        subject.check_heuristics(req)
      }.to not_change{entity.disabled}
        .and not_change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}
    end
  end

  context 'address blacklist expiration' do 
    it 'should do nothing' do
      Postfix::Rate::Models::NetworkAddress.create(address: '1.2.3.4', notes: 'salutava sempre...', inserted: Time.now, flags: 'b')
      req = request

      Timecop.travel((Postfix::Rate::Policyd::Heuristics::BLACKLIST_PERSISTENCE/2).seconds.from_now) do
        expect {
          expect(subject.check_blacklisting(req)).to be_nack
        }.to not_change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}
      end
    end

    it 'expire the blacklist' do
      Postfix::Rate::Models::NetworkAddress.create(address: '1.2.3.4', notes: 'salutava sempre...', inserted: Time.now, flags: 'b')
      req = request

      Timecop.travel((Postfix::Rate::Policyd::Heuristics::BLACKLIST_PERSISTENCE+10).seconds.from_now) do
        expect {
          expect(subject.check_blacklisting(req)).to be_nil
        }.to change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}.by(-1)
      end
    end
  end

  context 'policy with heuristics disabled' do
    let :identity do
      id = Postfix::Rate::Models::AuthenticatedIdentity.new(identity: 'sender@foo.bar')
      id.policy = Postfix::Rate::Models::Policy.find(name: 'default')
      id.save
      id
    end

    context 'should update the heuristics data' do
      it 'should initialize the heuristics data' do
        id = identity
        req = request
        entity = subject.fetch_entity(req)

        expect {
          subject.check_heuristics(req)
        }.to change{entity.tracker.heuristics}
      end

      it 'should add the new address in the heuristics data' do
        id = identity
        req = request
        entity = subject.fetch_entity(req)
        subject.update_quota(req)

        req1 = request
        req1.set_attribute 'client_address', '2.3.4.5'
        req1.new_instance!
        entity = subject.fetch_entity(req1)

        expect {
          subject.check_heuristics(req1)
        }.to change{entity.tracker.heuristics[:addresses].count}.to(2)
      end

      it 'should update the address data' do
        id = identity
        req = request
        entity = subject.fetch_entity(req)
        subject.update_quota(req)

        Timecop.travel(10.seconds.from_now) do
          req1 = request
          entity = subject.fetch_entity(req1)

          expect {
            subject.check_heuristics(req1)
          }.to change{entity.tracker.heuristics[:addresses]['1.2.3.4'][:last_seen]}
            .and change{entity.tracker.heuristics[:addresses]['1.2.3.4'][:total_count]}
            .and change{entity.tracker.heuristics[:addresses]['1.2.3.4'][:last_count]}
        end
      end

      it 'last_count is a rolling average' do
        id = identity
        req = request
        entity = subject.fetch_entity(req)
        subject.update_quota(req)

        Timecop.travel((entity.policy.h_interval/2).seconds.from_now) do
          req1 = request
          entity = subject.fetch_entity(req1)

          expect {
            subject.check_heuristics(req1)
          }.to change{entity.tracker.heuristics[:addresses]['1.2.3.4'][:last_count]}.to(1.5)
        end
      end

      it 'should reset last_count' do
        id = identity
        req = request
        entity = subject.fetch_entity(req)
        subject.update_quota(req)

        Timecop.travel((entity.policy.h_interval+10).seconds.from_now) do
          req1 = request
          entity = subject.fetch_entity(req1)
          
          entity.tracker.heuristics[:addresses]['1.2.3.4'][:last_count] = 100

          expect {
            subject.check_heuristics(req1)
          }.to change{entity.tracker.heuristics[:addresses]['1.2.3.4'][:last_count]}.to(1)
        end
      end
    end
    
    context 'should not ban the identity' do
      it 'when clients has multiple addresses and one address over quota' do
        id = identity
        req = request
        subject.fetch_entity(req)
        subject.update_quota(req)
        
        req1 = request
        req1.set_attribute 'client_address', '2.3.4.5'
        req1.new_instance!
        subject.fetch_entity(req1)
        subject.update_quota(req1)

        req2 = request
        req2.set_attribute 'client_address', '2.3.4.5'
        req2.new_instance!
        entity = subject.fetch_entity(req2)
        
        entity.tracker.heuristics[:addresses]['2.3.4.5'][:last_count] = entity.policy.h_interval_count
        entity.tracker.heuristics[:addresses]['2.3.4.5'][:total_count] = entity.policy.h_interval_count

        Timecop.travel(10.seconds.from_now) do
          expect {
            subject.check_heuristics(req2)
          }.to not_change{entity.disabled}
            .and not_change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}
        end
      end

      it 'when client is sending from too many addresses' do
        id = identity
        req = request
        entity = subject.fetch_entity(req)

        # build and store the maximum allowed amount of addresses
        addresses = 1.upto(entity.policy.h_max_addresses).map do | n |
          "1.1.1.#{n}"
        end

        addresses.each.with_index do | addr, idx |
          req = request
          req.set_attribute 'client_address', addr
          req.new_instance!
          entity = subject.fetch_entity(req)
          subject.update_quota(req)
        end

        # now attempt a request from a new address
        req = request
        req.set_attribute 'client_address', '2.2.2.2'
        req.new_instance!
        entity = subject.fetch_entity(req)
        
        # should forget 1.1.1.1 and store 2.2.2.2
        new_addresses = addresses.dup
        new_addresses.shift
        new_addresses.push('2.2.2.2')

        expect {
          subject.check_heuristics(req)
        }.to not_change{entity.disabled}
          .and not_change{Postfix::Rate::Models::NetworkAddress.where(flags: 'b').count}
          .and not_change{entity.tracker.heuristics[:addresses].keys}
      end
    end
  end
end