require 'postfix/rate/policyd/agent_backend'

request = %Q(request=smtpd_access_policy
protocol_state=RCPT
protocol_name=SMTP
helo_name=some.domain.tld
queue_id=8045F2AB23
sender=foo@bar.tld
recipient=bar@foo.tld
recipient_count=5
client_address=1.2.3.4
client_name=another.domain.tld
reverse_client_name=another.domain.tld
instance=123.456.7
sasl_method=plain
sasl_username=foobar@foo.bar
size=0

)

end_of_msg_request = %Q(request=smtpd_access_policy
protocol_state=END-OF-MESSAGE
protocol_name=SMTP
helo_name=some.domain.tld
queue_id=8045F2AB23
sender=foo@bar.tld
recipient=bar@foo.tld
recipient_count=5
client_address=1.2.3.4
client_name=another.domain.tld
reverse_client_name=another.domain.tld
instance=123.456.7
sasl_method=plain
sasl_username=foobar@foo.bar
size=1000

)

class MockTCPSocket
  attr_accessor :packet, :closed, :results
  def initialize(packet)
    @packet = packet.lines
    @results = []
  end

  def peeraddr(*)
    ["AF_INET", 10031, 'localhost', '127.0.0.1']
  end

  def readline
    unless @packet&.any?
      raise EOFError.new, 'EOF'
    end
    @packet.shift
  end

  def close
    @closed = true
  end

  def closed?
    @closed
  end

  def write(result)
    @results << result
  end
end

RSpec.describe Postfix::Rate::Policyd::RequestWorker do
  before :each do
    DatabaseHelpers.init_database
  end

  it 'should initialize' do
    expect{Postfix::Rate::Policyd::RequestWorker.new}.not_to raise_error
  end

  it 'should process_request' do
    subj = Postfix::Rate::Policyd::RequestWorker.new
    socket = MockTCPSocket.new request
    expect{subj.process_request(socket)}.not_to raise_error
  end

  it 'should close the socket' do
    subj = Postfix::Rate::Policyd::RequestWorker.new
    socket = MockTCPSocket.new request
    subj.process_request(socket)
    
    expect(socket.closed?).to be_truthy
  end

  it 'should send a response' do
    subj = Postfix::Rate::Policyd::RequestWorker.new
    socket = MockTCPSocket.new request
    subj.process_request(socket)
    
    expect(socket.results.count).to eq 1
  end

  it 'should process multiple requests' do
    subj = Postfix::Rate::Policyd::RequestWorker.new
    socket = MockTCPSocket.new request+request+request
    expect{subj.process_request(socket)}.not_to raise_error
    expect(socket.results.count).to eq 3
  end

  it 'should send a successful response' do
    subj = Postfix::Rate::Policyd::RequestWorker.new
    socket = MockTCPSocket.new request
    subj.process_request(socket)
    
    expect(socket.results.first).to eq "action=OK\n\n"
  end

  it 'should send a DEFER response if the identity is banned' do
    id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'foobar@foo.bar')
    id.disable! 'foo bar'

    subj = Postfix::Rate::Policyd::RequestWorker.new
    socket = MockTCPSocket.new request
    subj.process_request(socket)
    
    expect(socket.results.first).to match /\Aaction=DEFER /
  end

  it 'should send a DEFER response if the identity is over count quota' do
    id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'foobar@foo.bar')
    id.policy.count = 100
    id.policy.save
    id.tracker.count = id.policy.count
    id.tracker.save

    subj = Postfix::Rate::Policyd::RequestWorker.new
    socket = MockTCPSocket.new request
    subj.process_request(socket)
    
    expect(socket.results.first).to match /\Aaction=DEFER /
  end

  it 'should send a DEFER response if the identity is over recipients quota' do
    id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'foobar@foo.bar')
    id.policy.recipients = 100
    id.policy.save
    id.tracker.recipients = id.policy.recipients
    id.tracker.save

    subj = Postfix::Rate::Policyd::RequestWorker.new
    socket = MockTCPSocket.new request
    subj.process_request(socket)
    
    expect(socket.results.first).to match /\Aaction=DEFER /
  end

  it 'should send a DEFER response if the identity is over volume quota' do
    id = Postfix::Rate::Models::AuthenticatedIdentity.find_or_create(identity: 'foobar@foo.bar')
    id.policy.volume = 10000
    id.policy.save
    id.tracker.volume = id.policy.volume
    id.tracker.save

    subj = Postfix::Rate::Policyd::RequestWorker.new
    socket = MockTCPSocket.new end_of_msg_request
    subj.process_request(socket)
    
    expect(socket.results.first).to match /\Aaction=DEFER /
  end

  it 'should not raise exception when logging an exception' do
    subj = Postfix::Rate::Policyd::RequestWorker.new
    subj.backend = double('backend')
    allow(subj.backend).to receive(:check_initial_quota).and_raise(StandardError)
    allow(subj.backend).to receive(:check_rcpt_quota).and_raise(StandardError)
    allow(subj.backend).to receive(:check_final_quota_and_update).and_raise(StandardError)

    socket = MockTCPSocket.new end_of_msg_request

    # celluloid v1.8.0: /lib/celluloid/internals/logger.rb:71
    # @exception_handlers is nil
    # probably a bug
    # raises exception when calling crash()

    expect {
      subj.process_request(socket)
    }.not_to raise_error
    
    expect(socket.closed?).to be_truthy
    #expect(socket.results.first).to match /\Aaction=DEFER /
  end
end